import collections.abc
import typing as t
import unittest

from uritemplate import URITemplate
from uritemplate import expand
from uritemplate import partial
from uritemplate import variable
from uritemplate import variables


def merge_dicts(
    *args: t.Union[
        variable.VariableValueDict, t.Dict[str, str], t.Dict[str, t.List[str]]
    ]
) -> variable.VariableValueDict:
    d: t.Dict[str, variable.VariableValue] = {}
    for arg in args:
        d.update(arg)
    return d


ExampleVariables = variable.VariableValueDict
ExampleTemplatesAndResults = t.List[t.Tuple[str, t.Union[str, t.List[str]]]]


class ExampleWithVariables(t.TypedDict):
    expansion: ExampleVariables
    expected: str


class RFCTemplateExamples(type):
    var = {"var": "value"}
    hello = {"hello": "Hello World!"}
    path = {"path": "/foo/bar"}
    x = {"x": "1024"}
    y = {"y": "768"}
    empty = {"empty": ""}
    merged_x_y = merge_dicts(x, y)
    list_ex = {"list": ["red", "green", "blue"]}
    keys = {"keys": [("semi", ";"), ("dot", "."), ("comma", ",")]}

    # # Level 1
    # Simple string expansion
    level1_examples = {
        "{var}": {
            "expansion": var,
            "expected": "value",
        },
        "{hello}": {
            "expansion": hello,
            "expected": "Hello%20World%21",
        },
    }

    # # Level 2
    # Reserved string expansion
    level2_reserved_examples = {
        "{+var}": {
            "expansion": var,
            "expected": "value",
        },
        "{+hello}": {
            "expansion": hello,
            "expected": "Hello%20World!",
        },
        "{+path}/here": {
            "expansion": path,
            "expected": "/foo/bar/here",
        },
        "here?ref={+path}": {
            "expansion": path,
            "expected": "here?ref=/foo/bar",
        },
    }

    # Fragment expansion, crosshatch-prefixed
    level2_fragment_examples = {
        "X{#var}": {
            "expansion": var,
            "expected": "X#value",
        },
        "X{#hello}": {"expansion": hello, "expected": "X#Hello%20World!"},
    }

    # # Level 3
    # String expansion with multiple variables
    level3_multiple_variable_examples = {
        "map?{x,y}": {
            "expansion": merged_x_y,
            "expected": "map?1024,768",
        },
        "{x,hello,y}": {
            "expansion": merge_dicts(x, y, hello),
            "expected": "1024,Hello%20World%21,768",
        },
    }

    # Reserved expansion with multiple variables
    level3_reserved_examples = {
        "{+x,hello,y}": {
            "expansion": merge_dicts(x, y, hello),
            "expected": "1024,Hello%20World!,768",
        },
        "{+path,x}/here": {
            "expansion": merge_dicts(path, x),
            "expected": "/foo/bar,1024/here",
        },
    }

    # Fragment expansion with multiple variables
    level3_fragment_examples = {
        "{#x,hello,y}": {
            "expansion": merge_dicts(x, y, hello),
            "expected": "#1024,Hello%20World!,768",
        },
        "{#path,x}/here": {
            "expansion": merge_dicts(path, x),
            "expected": "#/foo/bar,1024/here",
        },
    }

    # Label expansion, dot-prefixed
    level3_label_examples = {
        "X{.var}": {
            "expansion": var,
            "expected": "X.value",
        },
        "X{.x,y}": {
            "expansion": merged_x_y,
            "expected": "X.1024.768",
        },
    }

    # Path segments, slash-prefixed
    level3_path_segment_examples = {
        "{/var}": {
            "expansion": var,
            "expected": "/value",
        },
        "{/var,x}/here": {
            "expansion": merge_dicts(var, x),
            "expected": "/value/1024/here",
        },
    }

    # Path-style parameters, semicolon-prefixed
    level3_path_semi_examples = {
        "{;x,y}": {
            "expansion": merged_x_y,
            "expected": ";x=1024;y=768",
        },
        "{;x,y,empty}": {
            "expansion": merge_dicts(x, y, empty),
            "expected": ";x=1024;y=768;empty",
        },
    }

    # Form-style query, ampersand-separated
    level3_form_amp_examples = {
        "{?x,y}": {
            "expansion": merged_x_y,
            "expected": "?x=1024&y=768",
        },
        "{?x,y,empty}": {
            "expansion": merge_dicts(x, y, empty),
            "expected": "?x=1024&y=768&empty=",
        },
    }

    # Form-style query continuation
    level3_form_cont_examples = {
        "?fixed=yes{&x}": {
            "expansion": x,
            "expected": "?fixed=yes&x=1024",
        },
        "{&x,y,empty}": {
            "expansion": merge_dicts(x, y, empty),
            "expected": "&x=1024&y=768&empty=",
        },
    }

    # # Level 4
    # String expansion with value modifiers
    level4_value_modifier_examples = {
        "{var:3}": {
            "expansion": var,
            "expected": "val",
        },
        "{var:30}": {
            "expansion": var,
            "expected": "value",
        },
        "{list}": {
            "expansion": list_ex,
            "expected": "red,green,blue",
        },
        "{list*}": {
            "expansion": list_ex,
            "expected": "red,green,blue",
        },
        "{keys}": {
            "expansion": keys,
            "expected": "semi,%3B,dot,.,comma,%2C",
        },
        "{keys*}": {
            "expansion": keys,
            "expected": "semi=%3B,dot=.,comma=%2C",
        },
    }

    # Reserved expansion with value modifiers
    level4_reserved_examples = {
        "{+path:6}/here": {
            "expansion": path,
            "expected": "/foo/b/here",
        },
        "{+list}": {
            "expansion": list_ex,
            "expected": "red,green,blue",
        },
        "{+list*}": {
            "expansion": list_ex,
            "expected": "red,green,blue",
        },
        "{+keys}": {
            "expansion": keys,
            "expected": "semi,;,dot,.,comma,,",
        },
        "{+keys*}": {
            "expansion": keys,
            "expected": "semi=;,dot=.,comma=,",
        },
    }

    # Fragment expansion with value modifiers
    level4_fragment_examples = {
        "{#path:6}/here": {
            "expansion": path,
            "expected": "#/foo/b/here",
        },
        "{#list}": {
            "expansion": list_ex,
            "expected": "#red,green,blue",
        },
        "{#list*}": {
            "expansion": list_ex,
            "expected": "#red,green,blue",
        },
        "{#keys}": {"expansion": keys, "expected": "#semi,;,dot,.,comma,,"},
        "{#keys*}": {"expansion": keys, "expected": "#semi=;,dot=.,comma=,"},
    }

    # Label expansion, dot-prefixed
    level4_label_examples = {
        "X{.var:3}": {
            "expansion": var,
            "expected": "X.val",
        },
        "X{.list}": {
            "expansion": list_ex,
            "expected": "X.red,green,blue",
        },
        "X{.list*}": {
            "expansion": list_ex,
            "expected": "X.red.green.blue",
        },
        "X{.keys}": {
            "expansion": keys,
            "expected": "X.semi,%3B,dot,.,comma,%2C",
        },
        "X{.keys*}": {
            "expansion": keys,
            "expected": "X.semi=%3B.dot=..comma=%2C",
        },
    }

    # Path segments, slash-prefixed
    level4_path_slash_examples = {
        "{/var:1,var}": {
            "expansion": var,
            "expected": "/v/value",
        },
        "{/list}": {
            "expansion": list_ex,
            "expected": "/red,green,blue",
        },
        "{/list*}": {
            "expansion": list_ex,
            "expected": "/red/green/blue",
        },
        "{/list*,path:4}": {
            "expansion": merge_dicts(list_ex, path),
            "expected": "/red/green/blue/%2Ffoo",
        },
        "{/keys}": {
            "expansion": keys,
            "expected": "/semi,%3B,dot,.,comma,%2C",
        },
        "{/keys*}": {
            "expansion": keys,
            "expected": "/semi=%3B/dot=./comma=%2C",
        },
    }

    # Path-style parameters, semicolon-prefixed
    level4_path_semi_examples = {
        "{;hello:5}": {
            "expansion": hello,
            "expected": ";hello=Hello",
        },
        "{;list}": {
            "expansion": list_ex,
            "expected": ";list=red,green,blue",
        },
        "{;list*}": {
            "expansion": list_ex,
            "expected": ";list=red;list=green;list=blue",
        },
        "{;keys}": {
            "expansion": keys,
            "expected": ";keys=semi,%3B,dot,.,comma,%2C",
        },
        "{;keys*}": {
            "expansion": keys,
            "expected": ";semi=%3B;dot=.;comma=%2C",
        },
    }

    # Form-style query, ampersand-separated
    level4_form_amp_examples = {
        "{?var:3}": {
            "expansion": var,
            "expected": "?var=val",
        },
        "{?list}": {
            "expansion": list_ex,
            "expected": "?list=red,green,blue",
        },
        "{?list*}": {
            "expansion": list_ex,
            "expected": "?list=red&list=green&list=blue",
        },
        "{?keys}": {
            "expansion": keys,
            "expected": "?keys=semi,%3B,dot,.,comma,%2C",
        },
        "{?keys*}": {
            "expansion": keys,
            "expected": "?semi=%3B&dot=.&comma=%2C",
        },
    }

    # Form-style query continuation
    level4_form_query_examples = {
        "{&var:3}": {
            "expansion": var,
            "expected": "&var=val",
        },
        "{&list}": {
            "expansion": list_ex,
            "expected": "&list=red,green,blue",
        },
        "{&list*}": {
            "expansion": list_ex,
            "expected": "&list=red&list=green&list=blue",
        },
        "{&keys}": {
            "expansion": keys,
            "expected": "&keys=semi,%3B,dot,.,comma,%2C",
        },
        "{&keys*}": {
            "expansion": keys,
            "expected": "&semi=%3B&dot=.&comma=%2C",
        },
    }

    def __new__(
        cls, name: str, bases: t.Sequence[t.Any], attrs: t.Dict[str, t.Any]
    ) -> t.Any:
        def make_test(
            d: t.Dict[str, ExampleWithVariables],
        ) -> collections.abc.Callable[["unittest.TestCase"], None]:
            def _test_(self: "unittest.TestCase") -> None:
                for k, v in d.items():
                    with self.subTest(template=k, **v):
                        t = URITemplate(k)
                        self.assertEqual(
                            t.expand(v["expansion"]), v["expected"]
                        )

            return _test_

        examples = [
            (n, getattr(RFCTemplateExamples, n))
            for n in dir(RFCTemplateExamples)
            if n.startswith("level")
        ]

        for name, value in examples:
            testname = "test_%s" % name
            attrs[testname] = make_test(value)

        return type.__new__(cls, name, bases, attrs)  # type: ignore


class TestURITemplate(unittest.TestCase, metaclass=RFCTemplateExamples):
    def test_no_variables_in_uri(self) -> None:
        """
        This test ensures that if there are no variables present, the
        template evaluates to itself.
        """
        uri = "https://api.github.com/users"
        t = URITemplate(uri)
        self.assertEqual(t.expand(), uri)
        self.assertEqual(t.expand(users="foo"), uri)

    def test_all_variables_parsed(self) -> None:
        """
        This test ensures that all variables are parsed.
        """
        uris = [
            "https://api.github.com",
            "https://api.github.com/users{/user}",
            "https://api.github.com/repos{/user}{/repo}",
            "https://api.github.com/repos{/user}{/repo}/issues{/issue}",
        ]

        for i, uri in enumerate(uris):
            t = URITemplate(uri)
            self.assertEqual(len(t.variables), i)

    def test_expand(self) -> None:
        """
        This test ensures that expansion works as expected.
        """
        # Single
        t = URITemplate("https://api.github.com/users{/user}")
        expanded = "https://api.github.com/users/sigmavirus24"
        self.assertEqual(t.expand(user="sigmavirus24"), expanded)
        v = t.variables[0]
        self.assertEqual(v.expand({"user": None}), {"/user": ""})

        # Multiple
        t = URITemplate("https://api.github.com/users{/user}{/repo}")
        expanded = "https://api.github.com/users/sigmavirus24/github3.py"
        self.assertEqual(
            t.expand({"repo": "github3.py"}, user="sigmavirus24"), expanded
        )

    def test_str_repr(self) -> None:
        uri = "https://api.github.com{/endpoint}"
        t = URITemplate(uri)
        self.assertEqual(str(t), uri)
        self.assertEqual(str(t.variables[0]), "/endpoint")
        self.assertEqual(repr(t), 'URITemplate("%s")' % uri)
        self.assertEqual(repr(t.variables[0]), "URIVariable(/endpoint)")

    def test_hash(self) -> None:
        uri = "https://api.github.com{/endpoint}"
        self.assertEqual(hash(URITemplate(uri)), hash(uri))

    def test_default_value(self) -> None:
        uri = "https://api.github.com/user{/user=sigmavirus24}"
        t = URITemplate(uri)
        self.assertEqual(
            t.expand(), "https://api.github.com/user/sigmavirus24"
        )
        self.assertEqual(
            t.expand(user="lukasa"), "https://api.github.com/user/lukasa"
        )

    def test_query_expansion(self) -> None:
        t = URITemplate("{foo}")
        self.assertEqual(
            t.variables[0]._query_expansion("foo", None, False, None), None
        )

    def test_label_path_expansion(self) -> None:
        t = URITemplate("{foo}")
        self.assertEqual(
            t.variables[0]._label_path_expansion("foo", None, False, None),
            None,
        )

    def test_label_path_expansion_explode_slash(self) -> None:
        t = URITemplate("{/foo*}")
        self.assertEqual(
            t.variables[0]._label_path_expansion("foo", [], True, None), None
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion("foo", [None], True, None),
            None,
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", [None, None], True, None
            ),
            None,
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion("foo", ["one"], True, None),
            "one",
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", ["one", "two"], True, None
            ),
            "one/two",
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", ["one", None, "two"], True, None
            ),
            "one/two",
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion("foo", [""], True, None), ""
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion("foo", ["", ""], True, None),
            "/",
        )

        self.assertEqual(
            t.variables[0]._label_path_expansion("foo", {}, True, None), None
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", {"one": ""}, True, None
            ),
            "one=",
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", {"one": "", "two": ""}, True, None
            ),
            "one=/two=",
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", {"one": None}, True, None
            ),
            None,
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", {"one": None, "two": "two"}, True, None
            ),
            "two=two",
        )
        self.assertEqual(
            t.variables[0]._label_path_expansion(
                "foo", {"one": None, "two": None}, True, None
            ),
            None,
        )

    def test_semi_path_expansion(self) -> None:
        t = URITemplate("{foo}")
        v = t.variables[0]
        self.assertEqual(
            v._semi_path_expansion("foo", None, False, False), None
        )
        t.variables[0].operator = variable.Operator.form_style_query
        self.assertEqual(
            v._semi_path_expansion("foo", ["bar", "bogus"], True, False),
            "foo=bar&foo=bogus",
        )

    def test_string_expansion(self) -> None:
        t = URITemplate("{foo}")
        self.assertEqual(
            t.variables[0]._string_expansion("foo", None, False, False), None
        )

    def test_hashability(self) -> None:
        t = URITemplate("{foo}")
        u = URITemplate("{foo}")
        d = {t: 1}
        d[u] += 1
        self.assertEqual(d, {t: 2})

    def test_no_mutate(self) -> None:
        args: variable.VariableValueDict = {}
        t = URITemplate("")
        t.expand(args, key=1)
        self.assertEqual(args, {})


class TestVariableModule(unittest.TestCase):
    def test_is_list_of_tuples(self) -> None:
        a_list = [(1, 2), (3, 4)]
        self.assertEqual(variable.is_list_of_tuples(a_list), (True, a_list))

        b_list = [1, 2, 3, 4]
        self.assertEqual(variable.is_list_of_tuples(b_list), (False, None))

    def test_list_test(self) -> None:
        a_list = [1, 2, 3, 4]
        self.assertEqual(variable.list_test(a_list), True)

        b_list = str([1, 2, 3, 4])
        self.assertEqual(variable.list_test(b_list), False)

    def test_list_of_tuples_test(self) -> None:
        a_list = [(1, 2), (3, 4)]
        self.assertEqual(variable.dict_test(a_list), False)

        d = dict(a_list)
        self.assertEqual(variable.dict_test(d), True)


class TestAPI(unittest.TestCase):
    uri = "https://api.github.com{/endpoint}"

    def test_expand(self) -> None:
        self.assertEqual(
            expand(self.uri, {"endpoint": "users"}),
            "https://api.github.com/users",
        )

    def test_partial(self) -> None:
        self.assertEqual(partial(self.uri), URITemplate(self.uri))
        uri = self.uri + "/sigmavirus24{/other}"
        self.assertEqual(
            partial(uri, endpoint="users"),
            URITemplate("https://api.github.com/users/sigmavirus24{/other}"),
        )

    def test_variables(self) -> None:
        self.assertEqual(
            variables(self.uri), URITemplate(self.uri).variable_names
        )


class TestNativeTypeSupport(unittest.TestCase):
    context: variable.VariableValueDict = {
        "zero": 0,
        "one": 1,
        "digits": list(range(10)),
        "a_float": 3.1415,
    }

    def test_expand(self) -> None:
        self.assertEqual(expand("{zero}", self.context), "0")
        self.assertEqual(expand("{one}", self.context), "1")
        self.assertEqual(
            expand("{digits}", self.context), "0,1,2,3,4,5,6,7,8,9"
        )
        self.assertEqual(
            expand("{?digits,one}", self.context),
            "?digits=0,1,2,3,4,5,6,7,8,9&one=1",
        )
        self.assertEqual(
            expand("{/digits}", self.context), "/0,1,2,3,4,5,6,7,8,9"
        )
        self.assertEqual(
            expand("{/digits*}", self.context), "/0/1/2/3/4/5/6/7/8/9"
        )
        # explicit testing for issue #53
        self.assertEqual(expand("{/zero}", self.context), "/0")
        self.assertEqual(expand("{/zero,a_float}", self.context), "/0/3.1415")


if __name__ == "__main__":
    unittest.main()



import json
import os.path
import typing as t

import uritemplate


def fixture_file_path(filename: str) -> str:
    absolute_dir = os.path.abspath(os.path.dirname(__file__))
    filename = filename + ".json"
    return os.path.join(absolute_dir, "fixtures", filename)


ExampleVariables = uritemplate.variable.VariableValueDict
ExampleTemplatesAndResults = t.List[t.Tuple[str, t.Union[str, t.List[str]]]]


class ExampleWithVariables(t.TypedDict):
    variables: ExampleVariables
    testcases: ExampleTemplatesAndResults


Examples = t.Dict[str, ExampleWithVariables]


def load_examples(filename: str) -> Examples:
    path = fixture_file_path(filename)
    with open(path, encoding="utf-8") as examples_file:
        examples = t.cast(Examples, json.load(examples_file))
    return examples


def expected_set(expected: t.Union[t.List[str], str]) -> t.Set[str]:
    if isinstance(expected, list):
        return set(expected)
    return {expected}


class FixtureMixin:
    examples: Examples

    def _get_test(
        self, section: str
    ) -> t.Tuple[ExampleVariables, ExampleTemplatesAndResults]:
        test = t.cast(ExampleWithVariables, self.examples.get(section, {}))
        return (
            test.get("variables", {}),
            test.get("testcases", []),
        )

    def _test(self, testname: str) -> None:
        variables, testcases = self._get_test(testname)
        for template, expected in testcases:
            expected_templates = expected_set(expected)
            expanded = uritemplate.expand(template, variables)
            assert expanded in expected_templates, (  # nosec
                f"expanded {template!r} with {variables!r} "
                f"and got {expanded!r} but expected one of "
                f"{expected_templates!r}"
            )


class TestSpecExamples(FixtureMixin):
    examples = load_examples("spec-examples")

    def test_level_1(self) -> None:
        """Check that uritemplate.expand matches Level 1 expectations."""
        self._test("Level 1 Examples")

    def test_level_2(self) -> None:
        """Check that uritemplate.expand matches Level 2 expectations."""
        self._test("Level 2 Examples")

    def test_level_3(self) -> None:
        """Check that uritemplate.expand matches Level 3 expectations."""
        self._test("Level 3 Examples")

    def test_level_4(self) -> None:
        """Check that uritemplate.expand matches Level 4 expectations."""
        self._test("Level 4 Examples")


class TestSpecExamplesByRFCSection(FixtureMixin):
    examples = load_examples("spec-examples-by-section")

    def test_variable_expansion(self) -> None:
        """Check variable expansion."""
        self._test("3.2.1 Variable Expansion")

    def test_simple_string_expansion(self) -> None:
        """Check simple string expansion."""
        self._test("3.2.2 Simple String Expansion")

    def test_reserved_expansion(self) -> None:
        """Check reserved expansion."""
        self._test("3.2.3 Reserved Expansion")

    def test_fragment_expansion(self) -> None:
        """Check fragment expansion."""
        self._test("3.2.4 Fragment Expansion")

    def test_dot_prefixed_label_expansion(self) -> None:
        """Check label expansion with dot-prefix."""
        self._test("3.2.5 Label Expansion with Dot-Prefix")

    def test_path_segment_expansion(self) -> None:
        """Check path segment expansion."""
        self._test("3.2.6 Path Segment Expansion")

    def test_path_style_parameter_expansion(self) -> None:
        """Check path-style param expansion."""
        self._test("3.2.7 Path-Style Parameter Expansion")

    def test_form_style_query_expansion(self) -> None:
        """Check form-style query expansion."""
        self._test("3.2.8 Form-Style Query Expansion")

    def test_form_style_query_cntinuation(self) -> None:
        """Check form-style query continuation."""
        self._test("3.2.9 Form-Style Query Continuation")


class TestExtendedTests(FixtureMixin):
    examples = load_examples("extended-tests")

    def test_additional_examples_1(self) -> None:
        """Check Additional Examples 1."""
        self._test("Additional Examples 1")

    def test_additional_examples_2(self) -> None:
        """Check Additional Examples 2."""
        self._test("Additional Examples 2")

    def test_additional_examples_3(self) -> None:
        """Check Additional Examples 3."""
        self._test("Additional Examples 3: Empty Variables")

    def test_additional_examples_4(self) -> None:
        """Check Additional Examples 4."""
        self._test("Additional Examples 4: Numeric Keys")

    def test_additional_examples_5(self) -> None:
        """Check Additional Examples 5."""
        self._test("Additional Examples 5: Explode Combinations")

    def test_additional_examples_6(self) -> None:
        """Check Additional Examples 6."""
        self._test("Additional Examples 6: Reserved Expansion")



# From: https://github.com/ActiveState/code/blob/master/recipes/Python/576696_OrderedSet_with_Weakrefs/  # noqa
import typing as t
import weakref


class Link:
    """Representation of one item in a doubly-linked list."""

    __slots__ = ("prev", "next", "key", "__weakref__")
    prev: "Link"
    next: "Link"
    key: str


class OrderedSet(t.MutableSet[str]):
    """A set that remembers the order in which items were added."""

    # Big-O running times for all methods are the same as for regular sets.
    # The internal self.__map dictionary maps keys to links in a doubly linked
    # list. The circular doubly linked list starts and ends with a sentinel
    # element. The sentinel element never gets deleted (this simplifies the
    # algorithm). The prev/next links are weakref proxies (to prevent circular
    # references). Individual links are kept alive by the hard reference in
    # self.__map. Those hard references disappear when a key is deleted from
    # an OrderedSet.

    def __init__(self, iterable: t.Optional[t.Iterable[str]] = None):
        self.__root = root = Link()  # sentinel node for doubly linked list
        root.prev = root.next = root
        self.__map: t.MutableMapping[str, Link] = {}  # key --> link
        if iterable is not None:
            self |= iterable  # type: ignore

    def __len__(self) -> int:
        return len(self.__map)

    def __contains__(self, key: object) -> bool:
        return key in self.__map

    def add(self, key: str) -> None:
        # Store new key in a new link at the end of the linked list
        if key not in self.__map:
            self.__map[key] = link = Link()
            root = self.__root
            last = root.prev
            link.prev, link.next, link.key = last, root, key
            last.next = root.prev = weakref.proxy(link)

    def discard(self, key: str) -> None:
        # Remove an existing item using self.__map to find the link which is
        # then removed by updating the links in the predecessor and successors.
        if key in self.__map:
            link = self.__map.pop(key)
            link.prev.next = link.next
            link.next.prev = link.prev

    def __iter__(self) -> t.Generator[str, None, None]:
        # Traverse the linked list in order.
        root = self.__root
        curr = root.next
        while curr is not root:
            yield curr.key
            curr = curr.next

    def __reversed__(self) -> t.Generator[str, None, None]:
        # Traverse the linked list in reverse order.
        root = self.__root
        curr = root.prev
        while curr is not root:
            yield curr.key
            curr = curr.prev

    def pop(self, last: bool = True) -> str:
        if not self:
            raise KeyError("set is empty")
        key = next(reversed(self)) if last else next(iter(self))
        self.discard(key)
        return key

    def __repr__(self) -> str:
        if not self:
            return f"{self.__class__.__name__}()"
        return f"{self.__class__.__name__}({list(self)!r})"

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        other = t.cast(t.Iterable[str], other)
        return not self.isdisjoint(other)
